package digi.recipeManager;

import java.io.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.Logger;
import java.util.regex.*;

import org.bukkit.*;
import org.bukkit.command.CommandSender;
import org.bukkit.configuration.file.*;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.inventory.ItemStack;
import org.bukkit.plugin.*;
import org.bukkit.plugin.java.JavaPlugin;
import org.bukkit.scheduler.BukkitScheduler;

import digi.recipeManager.data.*;

public class Main extends JavaPlugin
{
	// TODO: Note to self: remember to always change these when updating the respective files
	// TODO: also update lang_en.yml's version on each update
	private static final String						lastChangedConfig		= "1.21";
	private static final String						lastChangedReadme		= "1.21b";
	private static final String						lastChangedAliases		= "1.21";
	
	protected static Server							sv;
	protected static Logger							log;
	protected static BukkitScheduler				bs;
	protected static PluginManager					pm;
	protected static Plugin							plugin;
	protected static String							pluginVersion;
	protected static File							pluginFolder;
	
	protected static HashMap<Location, Double>		furnacesSmelting		= new HashMap<Location, Double>();
	
	protected static final int						recipePlaceholderId		= 31;
	protected static final short					recipePlaceholderData	= 1337;
	
	protected static final List<String>				recipeLoadErrors		= new ArrayList<String>();
	private static String							currentFile				= null;
	private static int								currentFileLine			= 0;
	
	private static final HashMap<String, String>	itemAliases				= new HashMap<String, String>();
	
	@Override
	public void onLoad()
	{
		plugin = this;
		log = getLogger();
		sv = getServer();
		bs = sv.getScheduler();
		pm = sv.getPluginManager();
		
		final PluginDescriptionFile pluginDesc = getDescription();
		pluginVersion = pluginDesc.getVersion();
		pluginFolder = getDataFolder();
		
		loadSettings();
		
		if(Config.ExistingRecipes.character == 'r')
		{
			// This code is used to detect special mod recipes
			// Mods' recipes usually start with 'X' in front or return a null recipe
			final Iterator<org.bukkit.inventory.Recipe> recipes = sv.recipeIterator();
			org.bukkit.inventory.Recipe recipe;
			ItemStack result;
			
			while(recipes.hasNext())
			{
				recipe = recipes.next();
				
				if(recipe == null)
					continue;
				
				result = recipe.getResult();
				
				if(result == null)
					continue;
				
				if(result.getType().toString().toLowerCase().charAt(0) != 'x')
					recipes.remove();
			}
		}
	}
	
	@Override
	public void onEnable()
	{
		getCommand("rmrecipes").setExecutor(new CmdRecipes());
		getCommand("rmcheck").setExecutor(new CmdCheck());
		getCommand("rmreload").setExecutor(new CmdReload());
		
		// Execute task after all plugins have loaded
		
		bs.scheduleSyncDelayedTask(this, new Runnable()
		{
			@Override
			public void run()
			{
				log.info("Loading recipes...");
				
				loadRecipes(false);
				new Econ();
				new Events();
				
				log.info("Done.");
			}
		});
	}
	
	@Override
	public void onDisable()
	{
		bs.cancelTasks(plugin);
		furnacesSmelting.clear();
		Recipes.reset();
		Recipes.removeCustomRecipes();
	}
	
	protected static void loadSettings()
	{
		if(pluginFolder.mkdirs())
			log.fine("Directory structure created.");
		
		// Load: config.yml
		
		if(!new File(pluginFolder + File.separator + "config.yml").exists())
		{
			plugin.saveResource("config.yml", false);
			log.info("config.yml file created, you should configure it!");
		}
		
		plugin.reloadConfig();
		
		try
		{
			final String version = plugin.getConfig().getString("lastchanged", null);
			
			if(version == null || !version.equals(lastChangedConfig))
				log.warning("config.yml has changed! You should delete it, use rmreload to re-generate it and then re-configure it, and then rmreload again.");
		}
		catch(final Exception e)
		{
			log.warning("Error reading config.yml's version! You should delete it to allow it to re-generate the newest version!");
		}
		
		Main.log.fine("config.yml settings:");
		
		for(final Config key : Config.values())
		{
			key.reload();
		}
		
		// Apply item return items
		
		net.minecraft.server.Item.WATER_BUCKET.a((Config.ReturnBuckets.bool ? net.minecraft.server.Item.BUCKET : null));
		net.minecraft.server.Item.LAVA_BUCKET.a((Config.ReturnBuckets.bool ? net.minecraft.server.Item.BUCKET : null));
		net.minecraft.server.Item.MILK_BUCKET.a((Config.ReturnBuckets.bool ? net.minecraft.server.Item.BUCKET : null));
		
		net.minecraft.server.Item.MUSHROOM_SOUP.a((Config.ReturnBowl.bool ? net.minecraft.server.Item.BOWL : null));
		
		net.minecraft.server.Item.POTION.a((Config.ReturnPotions.bool ? net.minecraft.server.Item.GLASS_BOTTLE : null));
		net.minecraft.server.Item.EXP_BOTTLE.a((Config.ReturnPotions.bool ? net.minecraft.server.Item.GLASS_BOTTLE : null));
		
		// Load: aliases.yml
		
		final File aliasesFile = new File(pluginFolder + File.separator + "aliases.yml");
		
		if(!aliasesFile.exists())
		{
			plugin.saveResource("aliases.yml", false);
			log.info("aliases.yml file created, you should read it!");
		}
		
		final FileConfiguration aliases = YamlConfiguration.loadConfiguration(aliasesFile);
		
		try
		{
			final String version = aliases.getString("lastchanged", null);
			
			if(version == null || !version.equals(lastChangedAliases))
				log.warning("aliases.yml has changed! You should delete it, use rmreload to re-generate it and then re-configure it, and then rmreload again.");
		}
		catch(final Exception e)
		{
			log.warning("Error reading aliases.yml's version! You should delete it to allow it to re-generate the newest version!");
		}
		
		itemAliases.clear();
		String itemStr;
		
		for(final Entry<String, Object> entry : aliases.getValues(false).entrySet())
		{
			if(entry.getKey().equals("lastchanged"))
				continue;
			
			itemStr = entry.getValue().toString().toUpperCase();
			
			if(!itemStr.contains(":"))
				itemStr = itemStr + ":*";
			
			itemAliases.put(entry.getKey().toUpperCase(), itemStr);
		}
		
		// readme.txt
		
		final File readmeFile = new File(pluginFolder + File.separator + "readme.txt");
		
		if(!readmeFile.exists())
		{
			plugin.saveResource("readme.txt", false);
			log.info("readme.txt file created, you should read it to learn how to make recipes!");
		}
		else
		{
			try
			{
				final BufferedReader reader = new BufferedReader(new FileReader(readmeFile));
				final String line = reader.readLine();
				reader.close();
				
				final Matcher match = Pattern.compile("v(.*?) ").matcher(line);
				boolean replace = true;
				
				if(match.find())
				{
					final String version = line.substring(match.start() + 1, match.end()).trim();
					
					if(version.equals(lastChangedReadme))
						replace = false;
				}
				
				if(replace)
				{
					plugin.saveResource("readme.txt", true);
					log.info("readme.txt has changed, overwrited with new one, you should re-read it!");
				}
			}
			catch(final Exception e)
			{
				e.printStackTrace();
			}
		}
		
		// Load: lang_*.yml
		
		Lang.load();
	}
	
	protected static boolean loadRecipes(final boolean simulation)
	{
		if(!simulation)
		{
			Recipes.reset();
			
			switch(Config.ExistingRecipes.character)
			{
				case 'r':
					Recipes.removeCustomRecipes();
					break;
				
				case 'c':
					sv.clearRecipes();
					break;
			}
		}
		
		final File dir = new File(pluginFolder + File.separator + "recipes");
		
		dir.mkdirs();
		
		loadDirectory(dir, simulation);
		
		if(recipeLoadErrors.size() > 0)
		{
			final StringBuilder errors = new StringBuilder();
			
			for(final String error : recipeLoadErrors)
			{
				errors.append(error + "\r\n");
			}
			
			log.severe("There were errors processing the files: \r\n\r\n" + errors + "\r\n\r\n");
			recipeLoadErrors.clear();
		}
		
		if(!simulation)
		{
			log.fine("Total loaded:");
			log.fine("  " + Recipes.craft.size() + " crafting recipes");
			log.fine("  " + Recipes.combine.size() + " combining recipes");
			log.fine("  " + Recipes.smelt.size() + " smelting recipes");
			log.fine("  " + Recipes.fuels.size() + " fuels");
			
			FurnaceTask.start(); // restart furnace task if any custom time smelt recipes exist
		}
		
		return (recipeLoadErrors.size() == 0);
	}
	
	private static void loadDirectory(final File dir, final boolean simulation)
	{
		log.fine("Loading '" + dir.getName() + "' folder...");
		
		String fileName;
		
		for(final File file : dir.listFiles())
		{
			fileName = file.getName();
			
			if(file.isDirectory())
			{
				if(fileName.equalsIgnoreCase("disabled"))
				{
					log.fine("Skipped the 'disabled' folder");
					continue;
				}
				
				if(Config.ExistingRecipes.character == 'n' && fileName.equalsIgnoreCase("default"))
				{
					log.fine("Skipped the 'default' folder (due to config's 'existing-recipes' set to 'nothing')");
					continue;
				}
				
				loadDirectory(file, simulation);
			}
			else
			{
				if(fileName.endsWith(".txt"))
				{
					try
					{
						loadDataFile(file, simulation);
					}
					catch(final Exception e)
					{
						log.severe("Failed to load '" + file.toString() + "', error:");
						e.printStackTrace();
					}
				}
			}
		}
	}
	
	private static void loadDataFile(final File file, final boolean simulation) throws Exception
	{
		final String fileName = file.getName(); //.getPath().replace("plugins" + File.separator + "RecipeManager" + File.separator + "recipes" + File.separator, "");
		
		log.fine("Loading '" + fileName + "' file...");
		
		final BufferedReader reader = new BufferedReader(new InputStreamReader(new DataInputStream(new FileInputStream(file))));
		final Recipe recipeData = new Recipe();
		String error;
		String line;
		currentFile = fileName;
		currentFileLine = 0;
		
		while((line = readLine(reader)) != null)
		{
			if((line = Main.processLine(line)) == null)
				continue;
			
			try
			{
				if((line = Recipes.processRecipeData(line, reader, recipeData)) == null)
					continue;
				
				error = null;
				
				if(line.startsWith("CRAFT"))
					error = Recipes.craftRecipe(line, reader, new Recipe(recipeData), simulation);
				else if(line.startsWith("COMBINE"))
					error = Recipes.combineRecipe(line, reader, new Recipe(recipeData), simulation);
				else if(line.startsWith("SMELT"))
					error = Recipes.smeltRecipe(line, reader, new Recipe(recipeData), simulation);
				else if(line.startsWith("FUEL"))
					error = Recipes.fuelRecipe(line, reader, new Recipe(recipeData), simulation);
				else
					error = "Line " + currentFileLine + " was skipped: \"" + line + "\"";
				
				if(error != null)
					recipeError(error);
			}
			catch(final Exception e)
			{
				e.printStackTrace();
			}
		}
		
		if(currentFileLine == 0)
			log.warning("Recipe file '" + fileName + "' is empty.");
		
		reader.close();
	}
	
	protected static String readLine(final BufferedReader reader) throws Exception
	{
		currentFileLine++; // to easily keep track of lines
		
		return reader.readLine();
	}
	
	protected static String processLine(String string)
	{
		if(string == null || string.isEmpty())
			return null;
		
		string = string.trim().toUpperCase();
		
		final String comments[] =
		{
			"//",
			"#"
		};
		
		int index;
		
		for(final String comm : comments)
		{
			index = string.indexOf(comm);
			
			if(index == 0)
				return null;
			
			if(index > -1)
				return string.substring(0, index).trim();
		}
		
		return string;
	}
	
	protected static Item processItem(String string, final int defaultData, final boolean allowData, final boolean allowAmount, final boolean allowEnchantments, final boolean printRecipeErrors)
	{
		string = string.trim();
		
		if(string.length() == 0)
			return null;
		
		final String itemString[] = string.split("\\|");
		final String stringArray[] = itemString[0].trim().split(":");
		
		Material mat = null;
		int type = 0;
		int data = defaultData;
		int amount = 1;
		
		if(stringArray.length <= 0 || stringArray[0].isEmpty())
			return new Item(0);
		
		stringArray[0] = stringArray[0].trim();
		final String alias = itemAliases.get(stringArray[0]);
		
		if(alias != null)
			return processItem(string.replace(stringArray[0], alias), defaultData, allowData, allowAmount, allowEnchantments, printRecipeErrors);
		
		mat = Material.matchMaterial(stringArray[0]);
		
		if(mat == null)
		{
			if(printRecipeErrors)
				recipeError("Item '" + stringArray[0] + "' does not exist! (TIP: name could be different, look in readme.txt for links)");
			
			return null;
		}
		
		type = mat.getId();
		
		if(type <= 0) // if it's AIR we don't need to move on
			return new Item(0);
		
		if(stringArray.length > 1)
		{
			if(allowData)
			{
				try
				{
					stringArray[1] = stringArray[1].trim();
					
					if(stringArray[1].charAt(0) != '*')
						data = Math.max(Integer.valueOf(stringArray[1]), data);
				}
				catch(final Exception e)
				{
					if(printRecipeErrors)
						recipeError("Item '" + mat + " has data value that is not a number: '" + stringArray[1] + "', defaulting to " + defaultData);
				}
			}
			else
			{
				if(printRecipeErrors)
					recipeError("Item '" + mat + "' can't have data value defined in this recipe's slot, data value ignored.");
			}
		}
		
		if(stringArray.length > 2)
		{
			if(allowAmount)
			{
				try
				{
					amount = Math.max(Integer.parseInt(stringArray[2].trim()), 1);
				}
				catch(final Exception e)
				{
					if(printRecipeErrors)
						recipeError("Item '" + mat + "' has amount value that is not a number: " + stringArray[2] + ", defaulting to 1");
				}
			}
			else
			{
				if(printRecipeErrors)
					recipeError("Item '" + mat + "' can't have amount defined in this recipe's slot, amount ignored.");
			}
		}
		
		final Item item = new Item(type, amount, (short)data);
		
		if(itemString.length > 1)
		{
			if(allowEnchantments)
			{
				if(item.amount > 1)
				{
					if(printRecipeErrors)
						recipeError("Item '" + mat + "' has enchantments and more than 1 amount, it can't have both, amount set to 1.");
					
					item.amount = 1;
				}
				
				final String[] enchants = itemString[1].split(",");
				String[] enchData;
				Enchantment ench;
				int level;
				int id;
				
				for(String enchant : enchants)
				{
					enchant = enchant.trim().toUpperCase();
					enchData = enchant.split(":");
					
					if(enchData.length != 2)
					{
						if(printRecipeErrors)
							recipeError("Enchantments have to be 'ENCHANTMENT:LEVEL' format. (TIP: see readme.txt for enchantment list link)");
						
						continue;
					}
					
					ench = Enchantment.getByName(enchData[0]);
					
					if(ench == null)
					{
						try
						{
							id = Integer.valueOf(enchData[0]);
							
							if(id <= 0)
								throw new Exception();
						}
						catch(final Exception e)
						{
							if(printRecipeErrors)
								recipeError("Enchantment '" + enchData[0] + "' does not exist! (TIP: name could be different, see readme.txt about enchantments list links)");
							
							continue;
						}
						
						ench = Enchantment.getById(id);
					}
					else
						id = ench.getId();
					
					if(!ench.canEnchantItem(item.getItemStack()))
					{
						if(printRecipeErrors)
							recipeError("Can't add enchantment '" + ench.toString() + "' on the '" + item + "' item!");
						
						continue;
					}
					
					if(enchData[1].equals("MAX"))
						level = ench.getMaxLevel();
					
					else
					{
						try
						{
							level = Integer.valueOf(enchData[1]);
							
							if(level <= 0)
								throw new Exception();
						}
						catch(final Exception e)
						{
							if(printRecipeErrors)
								recipeError("Invalid enchantment level: '" + enchData[1] + "' must be a number!");
							
							continue;
						}
						
						level = Math.min(Math.max(level, ench.getStartLevel()), ench.getMaxLevel());
					}
					
					item.enchantments.put(ench, level);
				}
			}
			else
			{
				if(printRecipeErrors)
					recipeError("Item '" + mat + "' can't use enchantments in this recipe slot!");
			}
		}
		
		return item;
	}
	
	public static boolean allowedToCraft(final CommandSender sender)
	{
		if(!sender.hasPermission("recipemanager.craft"))
		{
			if(Config.PrintCantCraft.bool)
				Main.printMessage(sender, Lang.get("notify.nopermissionall"));
			
			return false;
		}
		
		return true;
	}
	
	public static void printMessage(final CommandSender sender, final String[] messages)
	{
		for(String message : messages)
		{
			for(final ChatColor color : ChatColor.values())
			{
				message = message.replaceAll("(?i)<" + color.name() + ">", "" + color);
			}
		}
		
		sender.sendMessage(messages);
	}
	
	public static void printMessage(final CommandSender sender, String message)
	{
		for(final ChatColor color : ChatColor.values())
		{
			message = message.replaceAll("(?i)<" + color.name() + ">", "" + color);
		}
		
		sender.sendMessage(message);
	}
	
	protected static void recipeError(final String error)
	{
		recipeLoadErrors.add("(" + currentFile + ":" + currentFileLine + ") " + error);
	}
}