package digi.recipeManager;

import java.io.BufferedReader;
import java.util.*;
import java.util.Map.Entry;

import org.bukkit.inventory.*;
import org.bukkit.permissions.*;

import digi.recipeManager.data.*;
import digi.recipeManager.data.Recipe;

class Recipes
{
	protected static List<Craft>			craft				= new ArrayList<Craft>();
	protected static List<Combine>			combine				= new ArrayList<Combine>();
	protected static HashMap<String, Smelt>	smelt				= new HashMap<String, Smelt>();
	protected static HashMap<String, Fuel>	fuels				= new HashMap<String, Fuel>();
	
	protected static HashMap<String, Page>	playerPage			= new HashMap<String, Page>();
	
	private static int						craftNum			= 0;
	private static int						combineNum			= 0;
	protected static int					smeltCustomRecipes	= 0;
	
	protected static void reset()
	{
		fuels.clear();
		craft.clear();
		combine.clear();
		smelt.clear();
		
		craftNum = 0;
		combineNum = 0;
		smeltCustomRecipes = 0;
	}
	
	protected static String craftRecipe(String line, final BufferedReader reader, final Recipe recipeData, final boolean simulation) throws Exception
	{
		line = Main.processLine(Main.readLine(reader)); // skip recipe header, read next line
		line = Recipes.processRecipeData(line, reader, recipeData); // check for @flags
		
		if(line == null)
			return "Recipe has no ingredients !";
		
		final HashMap<ItemData, Character> itemChars = new HashMap<ItemData, Character>();
		final List<String> recipeShape = new ArrayList<String>();
		final ItemData[] ingredients = new ItemData[9];
		final Item[] leftovers = new Item[9];
		
		String split[];
//		String itemSplit[];
		String resultRaw = null;
		ItemData itemData;
		Character c;
		int charIndex = 97;
		int itemIndex = 0;
		int rows = 0;
		int largest = 0;
		boolean errors = false;
		
		while(rows < 3)
		{
			if(rows > 0)
				line = Main.processLine(Main.readLine(reader));
			
			if(line == null)
				continue;
			
			// if we bump into the result prematurely (smaller recipes)
			
			if(line.charAt(0) == '=')
			{
				resultRaw = line;
				break;
			}
			
			split = line.split("\\+");
			
			final int len = Math.min(split.length, 3); // max 3 items per row !
			final StringBuilder recipeChars = new StringBuilder();
			
			itemIndex = rows * 3;
			largest = Math.max(len, largest);
			
			for(int i = 0; i < len; i++)
			{
				// splitting line by ">" char to get leftover item (if any)
//				itemSplit = split[i].split(">");
				
				itemData = Main.processItem(split[i] /*itemSplit[0]*/, -1, true, false, false, true);
				
				if(itemData == null)
				{
					errors = true;
					continue;
				}
				
				if(itemData.type > 0)
				{
					final Character ic = itemChars.get(itemData);
					
					if(ic == null)
					{
						c = (char)charIndex++;
						
						itemChars.put(itemData, c);
					}
					else
						c = ic;
					
					ingredients[itemIndex] = itemData; // add non-AIR item
				}
				else
					c = ' ';
				
				recipeChars.append(c);
				
				// get leftover item (if any)
				/*
				if(itemData.type != 0 && itemSplit.length > 1)
					leftovers[itemIndex] = Main.processItem(itemSplit[1].trim(), 0, true, true);
				*/
				
				itemIndex++;
			}
			
			recipeShape.add(recipeChars.toString());
			rows++;
		}
		
		// If result wasn't found yet, it's got to be the 4th line
		
		if(resultRaw == null)
			resultRaw = Main.processLine(Main.readLine(reader));
		
		final List<Item> results = getResults(resultRaw, reader, false, false);
		
		if(results == null)
			return "Invalid result item(s)! (TIP: forgot the '=' ?)";
		
		if(errors)
			return "Recipe has some invalid ingredients, fix them!";
		
		// starting the recipe and setting it's result
		
		final ShapedRecipe recipe = new ShapedRecipe(new ItemStack(Main.recipePlaceholderId, craftNum, Main.recipePlaceholderData));
		final String[] chars = new String[recipeShape.size()];
		int i = 0;
		
		for(final String shapeRow : recipeShape)
		{
			chars[i++] = (shapeRow.length() < largest ? (shapeRow + "    ").substring(0, largest) : shapeRow);
		}
		
		recipe.shape(chars);
		
		for(final Entry<ItemData, Character> entry : itemChars.entrySet())
		{
			recipe.setIngredient(entry.getValue(), entry.getKey().getMaterial(), entry.getKey().data);
		}
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		for(final Craft r : craft)
		{
			if(r.ingredients == ingredients)
				return "Another recipe with the same ingredients already exists!";
		}
		
		// finally add the recipe to the server!
		
		if(!Main.sv.addRecipe(recipe))
			return "Couldn't add recipe, unknown error";
		
		craft.add(craftNum, new Craft(ingredients, leftovers, results, recipeData));
		craftNum++;
		
		return null;
	}
	
	protected static String combineRecipe(String line, final BufferedReader reader, final Recipe recipeData, final boolean simulation) throws Exception
	{
		line = Main.processLine(Main.readLine(reader)); // skip recipe header, read next line
		line = Recipes.processRecipeData(line, reader, recipeData); // check for @flags
		
		if(line == null)
			return "Recipe has no ingredients !";
		
		final String[] ingredientsRaw = line.split("\\+");
		
		final List<Item> results = getResults(Main.processLine(Main.readLine(reader)), reader, false, false);
		
		if(results == null)
			return "Invalid result item(s)! (TIP: forgot the '=' ?)";
		
		final ShapelessRecipe recipe = new ShapelessRecipe(new ItemStack(Main.recipePlaceholderId, combineNum, Main.recipePlaceholderData));
		final List<Item> ingredients = new ArrayList<Item>();
		Item item;
		int items = 0;
		
		for(final String itemRaw : ingredientsRaw)
		{
			item = Main.processItem(itemRaw, -1, true, true, false, true);
			
			if(item == null)
				return "Combine recipes can't have AIR ingredients !";
			
			if((items += item.amount) > 9)
				return "Combine recipes can't have more than 9 ingredients !";
			
			ingredients.add(item);
			recipe.addIngredient(item.amount, item.getMaterial(), item.data);
		}
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		for(final Combine r : combine)
		{
			if(r.ingredients == ingredients)
				return "Another recipe with the same ingredients already exists!";
		}
		
		if(!Main.sv.addRecipe(recipe))
			return "Couldn't add recipe, unknown error";
		
		combine.add(combineNum, new Combine(ingredients, results, recipeData));
		combineNum++;
		
		return null;
	}
	
	protected static String smeltRecipe(String line, final BufferedReader reader, final Recipe recipeData, final boolean simulation) throws Exception
	{
		line = Main.processLine(Main.readLine(reader)); // skip recipe header, read next line
		line = Recipes.processRecipeData(line, reader, recipeData); // check for @flags
		recipeData.noRewards(); // remove EXP, level and money costs or rewards
		
		if(line == null)
			return "Recipe has no ingredient !";
		
		final String[] split = line.split("%");
		
		if(split.length == 0)
			return "No ingredient!";
		
		final List<Item> results = getResults(Main.processLine(Main.readLine(reader)), reader, true, true);
		
		if(results == null)
			return "Invalid result item! (TIP: forgot the '=' ?)";
		
		if(results.size() > 1)
			Main.recipeError("Can't have more than 1 result in this recipe, the rest were ignored.");
		
		final Item result = results.get(0);
		
		// TODO: revert allowData to true when data values for furnaces have been fixed
		final Item ingredient = Main.processItem(split[0], -1, false, false, false, true);
		double minTime = -1.0;
		double maxTime = -1.0;
		
		if(split.length >= 2)
		{
			final String[] timeSplit = split[1].trim().split("-");
			
			if(!timeSplit[0].equals("INSTANT"))
				minTime = Double.valueOf(timeSplit[0]);
			
			if(timeSplit.length >= 2)
				maxTime = Double.valueOf(timeSplit[1]);
			
			if(maxTime > -1.0 && minTime >= maxTime)
				return "Smelting recipe has minimum time less or equal to maximum time! (TIP: Use a single number if you want a fixed value)";
		}
		
		// the recipe
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		if(Recipes.smelt.containsKey(ingredient.type + (ingredient.data == -1 ? "" : ":" + ingredient.data)))
			return "Recipe for smelting " + ingredient.printItemData() + " already exists! (TIP: if it's a MC recipe, edit it in the 'default' folder)";
		
		if(!Main.sv.addRecipe(new FurnaceRecipe((result.type == 0 ? new ItemStack(Main.recipePlaceholderId) : result.getItemStack()), ingredient.getMaterial(), ingredient.data)))
			return "Couldn't add recipe, unknown error";
		
		smelt.put(getIngredientString(ingredient), new Smelt(ingredient, result, minTime, maxTime, recipeData));
		
		if(minTime >= 0.0)
			smeltCustomRecipes++;
		
		return null;
	}
	
	protected static String fuelRecipe(String line, final BufferedReader reader, final Recipe recipeData, final boolean simulation) throws Exception
	{
		line = Main.processLine(Main.readLine(reader)); // skip recipe header, read next line
		line = Recipes.processRecipeData(line, reader, recipeData); // check for @flags
		recipeData.noRewards(); // remove EXP, level and money costs or rewards
		
		if(line == null)
			return "Recipe has no ingredient !";
		
		final String[] split = line.split("%");
		
		if(split[1] == null)
			return "No burn time set!";
		
		final String[] timeSplit = split[1].trim().split("-");
		final int minTime = Math.max(Integer.parseInt(timeSplit[0]), 1);
		int maxTime = -1;
		
		if(timeSplit.length >= 2)
			maxTime = Math.max(Integer.parseInt(timeSplit[1]), maxTime);
		
		final ItemData ingredient = Main.processItem(split[0], -1, true, false, false, true);
		
		if(ingredient == null || ingredient.type == 0)
			return "Invalid item: '" + ingredient + "'";
		
		if(minTime <= 0)
			return "Fuel " + ingredient.getMaterial() + ":" + ingredient.data + " can't burn for negative or zero seconds!";
		
		if(maxTime > -1 && minTime >= maxTime)
			return "Fuel " + ingredient.getMaterial() + ":" + ingredient.data + " has minimum time less or equal to maximum time! (TIP: Use a single number if you want a fixed value)";
		
		if(simulation) // if it's a simulation, no need to continue.
			return null;
		
		if(fuels.containsKey(ingredient))
			return "Fuel " + ingredient.getMaterial() + ":" + ingredient.data + " already exists!";
		
		fuels.put(getIngredientString(ingredient), new Fuel(ingredient, minTime, maxTime, recipeData));
		
		return null;
	}
	
	private static List<Item> getResults(String line, final BufferedReader reader, final boolean allowAir, final boolean oneResult) throws Exception
	{
		final List<Item> results = new ArrayList<Item>();
		Item noPercentItem = null;
		String[] resultRawSplit;
		Item item;
		int totalpercentage = 0;
		
		while(line != null && !line.isEmpty() && line.charAt(0) == '=')
		{
			resultRawSplit = line.substring(1).trim().split("%"); // remove the = character and see if it's got percentage
			
			if(resultRawSplit.length >= 2)
			{
				item = Main.processItem(resultRawSplit[1].trim(), 0, true, true, true, true);
				
				if(item == null || (!allowAir && item.type == 0))
				{
					Main.recipeError("Invalid result !");
					return null;
				}
				
				try
				{
					item.chance = Math.min(Math.max(Integer.valueOf(resultRawSplit[0].trim()), 1), 100);
				}
				catch(final Exception e)
				{
					Main.recipeError("Invalid percentage number: " + resultRawSplit[0]);
					return null;
				}
				
				if((totalpercentage += item.chance) > 100)
				{
					Main.recipeError("Total result items' chance exceeds 100% ! (TIP: not defining percentage for 1 item will make it's chance fit with the rest until 100%)");
					return null;
				}
				
				results.add(item);
			}
			else
			{
				if(resultRawSplit[0] == null)
					return null;
				
				if(noPercentItem != null)
				{
					Main.recipeError("Can't have more than 1 item without procentage to fill the rest!");
					return null;
				}
				
				noPercentItem = Main.processItem(resultRawSplit[0].trim(), 0, true, true, true, true);
				
				if(noPercentItem == null || (!allowAir && noPercentItem.type == 0))
				{
					Main.recipeError("Invalid result !");
					return null;
				}
			}
			
			line = Main.processLine(Main.readLine(reader));
		}
		
		if(noPercentItem != null)
		{
			noPercentItem.chance = (100 - totalpercentage);
			results.add(noPercentItem);
		}
		else if(results.isEmpty())
		{
			return null;
		}
		else if(!oneResult && totalpercentage < 100)
		{
			results.add(new Item(0, 0, (short)0, (100 - totalpercentage)));
		}
		
		return results;
	}
	
	private static String getIngredientString(final ItemData ingredient)
	{
		return ingredient.type + (ingredient.data == -1 ? "" : ":" + ingredient.data);
	}
	
	protected static Smelt getSmeltRecipe(final ItemStack ingredient)
	{
		final ItemData item = new ItemData(ingredient);
		final Smelt recipe = Recipes.smelt.get(item.type + "");
		
		if(recipe == null)
			return Recipes.smelt.get(item.type + ":" + item.data);
		
		return recipe;
	}
	
	protected static Fuel getFuelRecipe(final ItemStack ingredient)
	{
		final ItemData item = new ItemData(ingredient);
		final Fuel recipe = Recipes.fuels.get(item.type + "");
		
		if(recipe == null)
			return Recipes.fuels.get(item.type + ":" + item.data);
		
		return recipe;
	}
	
	protected static void removeCustomRecipes()
	{
		final Iterator<org.bukkit.inventory.Recipe> recipes = Main.sv.recipeIterator();
		org.bukkit.inventory.Recipe recipe;
		
		while(recipes.hasNext())
		{
			if((recipe = recipes.next()) != null && Recipes.isCustomRecipe(recipe.getResult()))
				recipes.remove();
		}
	}
	
	protected static boolean isCustomRecipe(final ItemStack result)
	{
		return (result != null && result.getTypeId() == Main.recipePlaceholderId && result.getDurability() == Main.recipePlaceholderData);
	}
	
	protected static List<Recipe> getRecipesForItem(final Item item, final boolean ingredient)
	{
		final List<Recipe> recipes = new ArrayList<Recipe>();
		
		for(final Craft r : craft)
		{
			if(ingredient)
			{
				for(final ItemData i : r.ingredients)
				{
					if(item.compareItemData(i))
					{
						recipes.add(r);
						break;
					}
				}
			}
			else
			{
				for(final Item i : r.results)
				{
					if(item.compareItemData(i))
					{
						recipes.add(r);
						break;
					}
				}
			}
		}
		
		for(final Combine r : combine)
		{
			for(final ItemData i : (ingredient ? r.ingredients : r.results))
			{
				if(item.compareItemData(i))
				{
					recipes.add(r);
					break;
				}
			}
		}
		
		for(final Smelt r : smelt.values())
		{
			if(item.compareItemData(ingredient ? r.ingredient : r.result))
				recipes.add(r);
		}
		
		if(ingredient)
		{
			for(final Fuel r : fuels.values())
			{
				if(item.compareItemData(r.ingredient))
					recipes.add(r);
			}
		}
		
		return recipes;
	}
	
	protected static String processRecipeData(String line, final BufferedReader reader, final Recipe recipeData) throws Exception
	{
		while(line != null && line.charAt(0) == '@')
		{
			line = line.substring(1).trim().toLowerCase(); // remove the @, any spaces and convert to lowercase
			
			recipeDataCheck(line, recipeData);
			
			line = Main.processLine(Main.readLine(reader));
		}
		
		return line;
	}
	
	private static void recipeDataCheck(final String line, final Recipe recipeData)
	{
		String[] split;
		
		if(line.startsWith("worlds"))
		{
			split = line.split(":");
			
			if(split.length < 2)
			{
				Main.recipeError("@worlds doesn't have any values! (TIP: @worlds: world, world2, etc)");
				return;
			}
			
			final String data = split[1].trim();
			
			if(data.equals("false"))
			{
				recipeData.worlds.clear();
				return;
			}
			
			split = data.split(",");
			
			for(String world : split)
			{
				world = world.trim();
				
				if(Main.sv.getWorld(world) == null)
				{
					Main.recipeError("@worlds has world that doesn't exist: '" + world + "'");
					continue;
				}
				
				recipeData.worlds.add(world);
			}
			
			return;
		}
		
		if(line.startsWith("permission"))
		{
			split = line.split(":");
			
			if(split.length < 1)
			{
				Main.recipeError("@permission doesn't have any permission node set! (TIP: @permission: permission.node)");
				return;
			}
			
			final String data = split[1].trim();
			
			if(data.equals("false"))
			{
				recipeData.permission = null;
				return;
			}
			
			final String[] args = data.split("=");
			final String perm = args[0].trim();
			
			if(perm.length() < 3 || !perm.matches("[a-zA-Z0-9.]+"))
			{
				Main.recipeError("@permission has invalid value: '" + perm + "', minimum 3 chars, numbers, letters and dots only!");
				return;
			}
			
			Permission p = Main.pm.getPermission(perm);
			final boolean exists = (p != null);
			
			if(!exists)
			{
				p = new Permission(perm);
				p.setDescription("RecipeManager's recipes");
			}
			
			if(args.length > 1)
			{
				final String defValue = args[1].trim();
				
				switch(defValue.charAt(0))
				{
					case 't':
					{
						p.setDefault(PermissionDefault.TRUE);
						break;
					}
					
					case 'f':
					{
						p.setDefault(PermissionDefault.FALSE);
						break;
					}
					
					case 'o':
					{
						p.setDefault(PermissionDefault.OP);
						break;
					}
					
					case 'n':
					{
						p.setDefault(PermissionDefault.NOT_OP);
						break;
					}
					
					default:
					{
						Main.recipeError("Invalid default value '" + defValue + "' for permission node '" + perm + "' (TIP: use only true/false/op/non-op)");
						return;
					}
				}
				
				Main.log.fine("Set permission " + perm + "'s default value to " + p.getDefault().toString());
			}
			
			recipeData.permission = p.getName();
			
			if(!exists)
				Main.pm.addPermission(p);
			
			return;
		}
		
		if(line.startsWith("givexp") || line.startsWith("giveexp") || line.startsWith("givelevel") || line.startsWith("givemoney"))
		{
			split = line.split(":");
			
			if(split.length < 2)
			{
				Main.recipeError("@" + split[0] + " doesn't a value!");
				return;
			}
			
			final String data = split[1].trim();
			int value = 0;
			
			if(!data.equals("false"))
			{
				try
				{
					value = Integer.valueOf(data);
				}
				catch(final Exception e)
				{
					Main.recipeError("@" + split[0] + " has invalid number value!");
					return;
				}
			}
			
			// TODO: investigate if these work!
			
			switch(line.charAt(4))
			{
				case 'e': // give[e]xp
				case 'x': // give[x]p
				{
					recipeData.giveExp = value;
					break;
				}
				
				case 'l': // give[l]evel
				{
					recipeData.giveLevel = value;
					break;
				}
				
				case 'm': // give[m]oney
				{
					recipeData.giveMoney = value;
					break;
				}
			}
			
			return;
		}
		
		if(line.startsWith("minxp") || line.startsWith("minexp") || line.startsWith("maxxp") || line.startsWith("maxexp") || line.startsWith("minlevel") || line.startsWith("maxlevel") || line.startsWith("minmoney") || line.startsWith("maxmoney"))
		{
			split = line.split(":");
			
			if(split.length < 2)
			{
				Main.recipeError("@" + split[0] + " doesn't a value!");
				return;
			}
			
			final String data = split[1].trim();
			int value = 0;
			
			if(!data.equals("false"))
			{
				try
				{
					value = Integer.valueOf(data);
				}
				catch(final Exception e)
				{
					Main.recipeError("@" + split[0] + " has invalid number value!");
					return;
				}
			}
			
			switch(line.charAt(3))
			{
				case 'e': // min/max[e]xp
				case 'x': // min/max[x]p
				{
					switch(line.charAt(2))
					{
						case 'n':
							recipeData.minExp = value;
							break;
						case 'x':
							recipeData.maxExp = value;
							break;
					}
					
					break;
				}
				
				case 'l': // min/max[l]evel
				{
					switch(line.charAt(2))
					{
						case 'n':
							recipeData.minLevel = value;
							break;
						case 'x':
							recipeData.maxLevel = value;
							break;
					}
					
					break;
				}
				
				case 'm': // min/max[m]oney
				{
					switch(line.charAt(2))
					{
						case 'n':
							recipeData.minMoney = value;
							break;
						case 'x':
							recipeData.maxMoney = value;
							break;
					}
					
					break;
				}
			}
			
			return;
		}
		
		Main.log.warning("Unknown @flag: '" + line + "' !");
	}
}