package digi.recipeManager.data;

import java.util.*;

import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;

import digi.recipeManager.*;

public class Recipe
{
	public String		permission	= null;
	public List<String>	worlds		= new ArrayList<String>();
	public int			giveExp		= 0;
	public int			minExp		= 0;
	public int			maxExp		= 0;
	public int			giveLevel	= 0;
	public int			minLevel	= 0;
	public int			maxLevel	= 0;
	public double		giveMoney	= 0;
	public double		minMoney	= 0;
	public double		maxMoney	= 0;
	
	public Recipe()
	{
	}
	
	public Recipe(final Recipe recipeData)
	{
		permission = recipeData.permission;
		worlds = recipeData.worlds;
		giveExp = recipeData.giveExp;
		minExp = recipeData.minExp;
		maxExp = recipeData.maxExp;
		giveLevel = recipeData.giveLevel;
		minLevel = recipeData.minLevel;
		maxLevel = recipeData.maxLevel;
		giveMoney = recipeData.giveMoney;
		minMoney = recipeData.minMoney;
		maxMoney = recipeData.maxMoney;
	}
	
	public boolean usableWorld(final String worldName)
	{
		return (worlds.size() > 0 ? worlds.contains(worldName.toLowerCase()) : true);
	}
	
	public void noRewards()
	{
		giveExp = 0;
		giveLevel = 0;
		giveMoney = 0;
	}
	
	public boolean usableBy(final CommandSender sender, final boolean printMessages)
	{
		if(permission != null && !sender.hasPermission(permission))
		{
			if(printMessages && Config.PrintNoPermission.bool)
				Main.printMessage(sender, Lang.get("notify.nopermission"));
			
			return false;
		}
		
		if(sender instanceof Player)
		{
			final Player player = (Player)sender;
			
			if(!sender.hasPermission("recipemanager.noworld") && !usableWorld(player.getWorld().getName()))
			{
				if(printMessages && Config.PrintNoWorld.bool)
					Main.printMessage(player, Lang.get("notify.noworld"));
				
				return false;
			}
			
			final int playerExp = Exp.getCurrentExp(player);
			
			if(minExp > 0 && minExp > playerExp)
			{
				if(printMessages && Config.PrintRequires.bool)
					Main.printMessage(player, Lang.get("notify.atleast", minExp + " " + Lang.get("general.exp")));
				
				return false;
			}
			
			if(maxExp > 0 && maxExp < playerExp)
			{
				if(printMessages && Config.PrintTooMuch.bool)
					Main.printMessage(player, Lang.get("notify.maximum", maxExp + " " + Lang.get("general.exp")));
				
				return false;
			}
			
			if(giveExp != 0 && (playerExp + giveExp) <= 0)
			{
				if(printMessages && Config.PrintNotEnough.bool)
					Main.printMessage(player, Lang.get("notify.notenough", Math.abs(giveExp) + " " + Lang.get("general.exp")));
				
				return false;
			}
			
			final int playerLevel = Exp.getCurrentLevel(playerExp);
			
			if(minLevel > 0 && minLevel > playerLevel)
			{
				if(printMessages && Config.PrintRequires.bool)
					Main.printMessage(player, Lang.get("notify.atleast", Lang.get("general.level") + " " + minLevel));
				
				return false;
			}
			
			if(maxLevel > 0 && maxLevel < playerLevel)
			{
				if(printMessages && Config.PrintTooMuch.bool)
					Main.printMessage(player, Lang.get("notify.maximum", Lang.get("general.level") + " " + maxLevel));
				
				return false;
			}
			
			if(giveLevel != 0 && (playerLevel + giveLevel) <= 0)
			{
				if(printMessages && Config.PrintNotEnough.bool)
					Main.printMessage(player, Lang.get("notify.notenough", Math.abs(giveLevel) + " " + Lang.get("general.level" + (Math.abs(giveLevel) == 1 ? "" : "s"))));
				
				return false;
			}
			
			if(Econ.enabled)
			{
				final double playerMoney = Econ.moneyAmount(player.getName());
				
				if(minMoney > 0 && minMoney > playerMoney)
				{
					if(printMessages && Config.PrintRequires.bool)
						Main.printMessage(player, Lang.get("notify.atleast", Econ.moneyFormat(minMoney)));
					
					return false;
				}
				
				if(maxMoney > 0 && maxMoney < playerMoney)
				{
					if(printMessages && Config.PrintTooMuch.bool)
						Main.printMessage(player, Lang.get("notify.maximum", Econ.moneyFormat(maxMoney)));
					
					return false;
				}
				
				if(giveMoney != 0 && (playerMoney + giveMoney) <= 0)
				{
					if(printMessages && Config.PrintNotEnough.bool)
						Main.printMessage(player, Lang.get("notify.notenough", Econ.moneyFormat(Math.abs(giveMoney))));
					
					return false;
				}
			}
		}
		
		return true;
	}
	
	public String[] print()
	{
		return new String[]
		{
			""
		};
	}
	
	public void affectExp(final Player player)
	{
		if(giveExp == 0)
			return;
		
		Exp.changeExp(player, giveExp);
		
		if(Config.PrintLost.bool)
			Main.printMessage(player, Lang.get("notify." + (giveExp > 0 ? "got" : "lost"), Math.abs(giveExp) + " " + Lang.get("general.exp")));
	}
	
	public void affectLevel(final Player player)
	{
		if(giveLevel == 0)
			return;
		
		player.setLevel(player.getLevel() + giveLevel);
		
		if(Config.PrintLost.bool)
			Main.printMessage(player, Lang.get("notify." + (giveLevel > 0 ? "got" : "lost"), Math.abs(giveLevel) + " " + Lang.get("general.level" + (Math.abs(giveLevel) == 1 ? "" : "s"))));
	}
	
	public void affectMoney(final Player player)
	{
		if(!Econ.enabled || giveMoney == 0)
			return;
		
		if(giveMoney > 0)
			Econ.moneyAdd(player.getName(), giveMoney);
		else
			Econ.moneySubtract(player.getName(), Math.abs(giveMoney));
		
		if(Config.PrintLost.bool)
			Main.printMessage(player, Lang.get("notify." + (giveMoney > 0 ? "got" : "lost"), Econ.moneyFormat(Math.abs(giveMoney))));
	}
}
